#Copyright 2008 Cold Spring Harbor Laboratory.
#Written by Yaniv Erlich (erlich@cshl.edu)

package Cluster::Blat;
require Exporter;

our @ISA    = ("Exporter");
our @EXPORT = qw(split_file);
our $VERSION = 1.00;


use strict;
use warnings;
use Readonly;
use Carp;
use List::MoreUtils qw (any);
use SGE_API;
use ExtraUtils qw(timestamp_file is_fasta);
use Submit_tools;
use File::Basename;
use Storable;
use Archive::Tar;
use Cwd qw(abs_path);
use FindBin qw($Bin);

use lib "$Bin";




Readonly my $D_BLAT_SCORE                   => '20';
Readonly my $D_BLAT_IDEN                    => '60';
Readonly my $D_BLAT_MIN_MATCH               => '1';
Readonly my $D_BLAT_TILE_SIZE               => '6';
Readonly my $D_BLAT_COMMAND                 => 'blat';
Readonly my $D_SUFFIX_TO_UNI_HASH           => 'PosToSeq.uniHash';


Readonly my $SPLIT_LINES_BLAT         => '10000';
Readonly my $D_BLAT_LIFETIME          => '10h';
Readonly my $D_REFRESH_TIME           => '0.2m';
Readonly my $D_ANALYZE_BLAT_LOCATION  => "$Bin/Cluster/Remote_Scripts/analyze_blat_neat.pl";

Readonly my $SPACE                    => q{ };



sub quick {

    my ($arg_ref) = @_;
    my $genome_reference = $arg_ref->{genome_reference_2bit}                ;
    my $genome_locality  = $arg_ref->{genome_locality}                      ;
    my $file_to_blat     = $arg_ref->{file_to_blat}                         ;
    my $minMatch         = $arg_ref->{minMatch}         || $D_BLAT_MIN_MATCH;
    my $tileSize         = $arg_ref->{tileSize}         || $D_BLAT_TILE_SIZE;
    my $minScore         = $arg_ref->{minScore}         || $D_BLAT_SCORE    ;
    my $minIdentity      = $arg_ref->{minIdentity}      || $D_BLAT_IDEN     ;
    my $dots             = $arg_ref->{dots}                                 ;
    my $verbose          = $arg_ref->{verbose}                              ;
    my $threshold_for_anl= $arg_ref->{threshold_for_analysis}               ;
    my $number_of_cycles = $arg_ref->{number_of_cycles}                     ;
    my $debug            = $arg_ref->{debug}                                ;

    confess "genome_locality must be either 'local' or 'distal'\n"
        if (($genome_locality ne 'local') and ($genome_locality ne 'distal'));

    confess "blat error\n" 
        if any {not defined $_} $genome_reference,
                                $file_to_blat    ,
                                $threshold_for_anl,
                                $number_of_cycles,
                                $genome_locality;




    $genome_reference = abs_path $genome_reference;
    $file_to_blat     = abs_path $file_to_blat;

    confess "Threshold_for_analysis error\n"
        if (
            ($threshold_for_anl < $minMatch)        or
            ($threshold_for_anl > $number_of_cycles)
           );

    #let's start!
    my $advance_parameters = "-minMatch=$minMatch ".
                             "-tileSize=$tileSize ".
                             "-minScore=$minScore ".
                             "-minIdentity=$minIdentity ".
                             "-out=pslx ";

    if (defined $dots) {
        $advance_parameters .= "-dots=$dots ";
    }

    if ($verbose) {

        my @param = split /\s+/, $advance_parameters;

        print "Sending $file_to_blat to blat on the cluster\n",
              "Your blat parameters are:\n";
               foreach my $p (@param) {
                 print $p,"\n";
               }
        print "Your refernce genome file is $genome_reference\n";
    }

    my $file_pos_to_hash = 

    cluster_blat_and_analyze ({genome_reference_2bit=>$genome_reference,
                               genome_locality=>$genome_locality,
                               input=>$file_to_blat,
                               advance_parameters=>$advance_parameters,
                               threshold_for_analysis=>$threshold_for_anl,
                               number_of_cycles=>$number_of_cycles,
                               verbose=>$verbose,
                               debug=>$debug,
                              });

    return $file_pos_to_hash;

}




 sub cluster_blat_and_analyze {

    my ($arg_ref) = @_;

    my $genome_reference = $arg_ref->{genome_reference_2bit};
    my $genome_locality = $arg_ref->{genome_locality};
    my $file_to_blat = $arg_ref->{input};
    my $number_of_cycles = $arg_ref->{number_of_cycles};
    my $advance_parameters = $arg_ref->{advance_parameters};
    my $threshold_for_analysis = $arg_ref->{threshold_for_analysis};
    my $verbose = $arg_ref->{verbose};
    my $debug = $arg_ref->{debug};

    confess "cluster_blat_and_analyze error\n"
        if any {not defined $_} $genome_reference,
                                $file_to_blat,
                                $threshold_for_analysis,
                                $advance_parameters;

    my $r_split_files = split_file({input => $file_to_blat,
                                    lines => $SPLIT_LINES_BLAT
                                    });


    my $number_of_files = scalar @$r_split_files;

    print "Split $file_to_blat to $number_of_files subfiles. Each contains ",
          "$SPLIT_LINES_BLAT lines\n"
            if ($verbose);

    print "\n",
          "--------------------------------------\n",
          "Names of split files:\n" if ($debug);

    foreach my $f (@$r_split_files) {
          print "$f\n" if ($debug);
          return if (not is_fasta($f));
    }

    print "--------------------------------------\n",
          "\n" if ($debug);



    my @starts =  1;
    my @gaps   =  1;
    my @stops  =  $number_of_files;



    my $cluster_script = SGE_API->new();
    my ($counter,undef) = $cluster_script->set_range(\@starts,\@stops,\@gaps);

    print "Total number of permutations is $counter\n" if ($debug);


    $cluster_script->add_required_file({refer=> 'genome', 
                                 name => $genome_reference, 
                                 place=> $genome_locality});

    $cluster_script->add_required_file({refer=>'file_to_blat',
                                 name => "$file_to_blat.<_>",
                                 place=> 'local'
                               });

    $cluster_script->add_command("$D_BLAT_COMMAND <genome> <file_to_blat> <file_to_blat>.blt $advance_parameters");


    $cluster_script->add_required_file({refer  =>'analyze_blat',
                                 name    =>$D_ANALYZE_BLAT_LOCATION,
                                 place   =>'local'
                                });

    $cluster_script->add_command("./<analyze_blat> <file_to_blat>.blt $threshold_for_analysis $number_of_cycles <genome> <file_to_blat>.hash");
    $cluster_script->add_command("mv <file_to_blat>.hash \$SOURCE_DIR");

    my $script_to_submit = $cluster_script->compile;
    $cluster_script->check_integrity;

    print "Generating script to send to the cluster: $script_to_submit\n" if ($debug);
    if ($debug) {_print_script($script_to_submit)};


    my $qsub = Submit_tools->new();
    $qsub->life_time($D_BLAT_LIFETIME);
    $qsub->range(1,$number_of_files,1);
    $qsub->submit($script_to_submit);
    my $job_name = $qsub->get_job_id;
    my @stdout_files = $qsub->get_message_filenames;
    $qsub->automatic_pilot($D_REFRESH_TIME);
    $cluster_script->bye;
    stdout_clean(\@stdout_files,$job_name,$verbose,$debug);


    my ($number_of_sequences_that_passed_analysis,
        $file_pos_to_hash) =   unify_hashes ($r_split_files,$file_to_blat);

    print "$number_of_sequences_that_passed_analysis sequences passed the BLAT and the analysis\n",
          "The pos_to_seq hash file is stored at $file_pos_to_hash\n";




    return $file_pos_to_hash;
}



sub unify_hashes {
    my ($r_split_files,$file_to_blat) = @_;
    my %seq_to_pos;
     foreach my $file (@{$r_split_files}) {

         my $hash_file = basename ($file) . '.hash';
         my $r_htmp    = retrieve ($hash_file);

         while (my ($k,$v) = each %$r_htmp) {$seq_to_pos{$k} = $v};
         unlink $file      or confess "can't clean $file: $!\n";
         unlink $hash_file or confess "can't clean $hash_file: $!\n";
     }
    my $outputfile = "$file_to_blat.$D_SUFFIX_TO_UNI_HASH";
    store (\%seq_to_pos,$outputfile);
    my $total_sequences_that_passed_analysis = scalar keys %seq_to_pos;
    return ($total_sequences_that_passed_analysis,$outputfile);
}



sub split_file {
    #############################
    #this function takes a file and split it into sub files. 
    # the input of the function is an hash table with the following keys: input, input_path, output_path
    # ,number_of_lines. input and number_of_lines are mandatory. the defaults of input_path and output_path
    # are the current directory.
    # The number of lines determines the size of each file, and it must be an even number for minimizing 
    # the risk of mistakes when dealing with FASTA formats or Solexa ouptput files.
    # The names of the output files will have '.' and a counter (For instance: x.1 x.2 x.3 ...)
    # The function returns a reference for all the new files that were created including the path.
    # Example:
    # $r_array = split_file({input=>'Utils.pm', output_path=>'./tmp', lines=>20, input_path=>'./'}).
    # will split ./Utils.pm into 20 lines files in ./tmp/Utils.pm.X. And it will return a
    # reference an for array with all the new filenames: (./tmp/Utils.pm.1, ./tmp/Utils.pm.2, ...).
    ##############################


    #get the function parameters
    my ($arg_ref) = @_;

    my $path_and_file_to_split = $arg_ref->{'input'}                                    ;
    my ($file_to_split,$input_path) = fileparse $path_and_file_to_split                 ;
    my $output_path   = $arg_ref->{'output_path'}   || $input_path                      ;
    my $lines         = $arg_ref->{'lines'}                                             ;

    #set variables:
    my $count       = 0; #counts number of lines
    my $file_count  = 0; #running counter for the suffix of the new files
    my $OUT         = undef; #filehandle of the output files
    my @output_files;        #holds the name of the output files including path

    #catching errors:
    confess "split_file error\n" 
        if any {not defined $_} $file_to_split, $lines;
    confess "number of lines in split_file function must be even\n"
        if ($lines % 2 != 0);
            #if number of lines is not even - quit. this is for extra protection for solexa outputs of fasta format.


    open my $IN, '<', $path_and_file_to_split
        or confess "can't read $file_to_split\n";

    while (<$IN>) {

        # the splitting criteria:
        if ($count % $lines == 0) {
            close $OUT if (defined $OUT);

            $file_count++;
                #count creation of a new file
            my $suffix      = $file_count; #Once it was with sprintf "%08d", 
            my $output_file = "$output_path/$file_to_split" . ".$suffix";
            push @output_files , $output_file;

            open $OUT , '>', $output_file 
                or confess "can't open $output_file for writing in split_file function\n";

        }
        print {$OUT} "$_"
            or confess "can't print to output file in split_file function\n";
            # here is actually print the input to the output.
        $count++;
            #count the new line
    }
    close $IN;
    return \@output_files;
}


sub stdout_clean {

    my ($r_files,$job_name,$verbose,$debug) = @_;
    my @files = map {abs_path $_ } @$r_files;

    if ($verbose) {
        print "Cleaning stdout files\n";
    }
    #if debug mode keep the files:
    if ($debug) {
        my $garb_dir_fullname = _make_grabage_directory($job_name);
        my $tar = Archive::Tar->new();
        $tar->add_files(@files);
        $tar->write("$garb_dir_fullname/$job_name.tar",1);
        print "files were tarred in $garb_dir_fullname under $job_name.tar";
    }
    #any case - clean the uncompressed files:
    foreach my $file (@files) {
        unlink $file ||
            croak "can't clean file: $file $!\n";
        if ($verbose) {print "$file cleaned\n"};
    }
}


sub _make_grabage_directory {
    my ($subdir) = @_;
    if (not -d './tmp') {
        mkdir 'tmp'
            or croak "can't make tmp dir:$!\n";
    }
    my $subdir_c = $subdir;
    while (-d "./tmp/$subdir_c") {
        $subdir_c = $subdir . time();
        carp "chaing name of tmp subdir to $subdir_c\n";
    }
    mkdir "./tmp/$subdir_c"
        or croak "can't make tmp subdir $subdir_c: $!\n";

    $subdir_c = abs_path "./tmp/$subdir_c";
    return $subdir_c;
}


sub test {

            quick
                ({
                    genome_reference_2bit => '/home2/hannon/erlich/SGE-API/step0/tmp/phiX.2bit',
                    genome_locality       => 'local',
                    file_to_blat          => '/home2/hannon/erlich/SGE-API/step0/tmp/s_8_0001_seq.txt.NoN.fa',
                    dots                  => 100,
                    verbose               => 1,
                    threshold_for_analysis=> 20,
                    number_of_cycles      => 50,
                    debug                 => 1,

                })

}


sub _print_script {

    my $filename = shift;
    print "\n",
          "-------------------------------------------\n",
    open my $FH,'<', $filename
        or confess "can't open $filename: $!\n";

    while (<$FH>) {
        print "$_";
    }
    print "-------------------------------------------\n",
          "\n";
    close $FH;
    return 1;
}



1;
