#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

#This package samples Solexa data (intensity values and sequences) and creates the training set.

package Snapshot;

require Exporter;

our @ISA    = ("Exporter");
our @EXPORT = qw(quick $D_SUFFIX_OF_INT_FILES $D_SYMBOL);
our $VERSION = 1.00;


use strict;
use warnings;
use Readonly;
use List::Util qw (shuffle);
use List::MoreUtils qw (any);
use Storable;
use Carp;
use FindBin;
use lib $FindBin::Bin;
use File::Basename;
use File::Copy;



Readonly my $D_FOLDER                               => 'Training_set_files';
Readonly my $D_NAME_OF_UNIFIED_SAMPLE_SEQ           => 'unified_seq_sample.txt';
Readonly my $D_NAME_OF_UNIFIED_SAMPLE_INT           => 'unified_int_sample.txt';
Readonly my $D_NAME_OF_UNIFIED_SAMPLE_SEQ_FOR_BLAT  => 'unified_seq_sample_for_blat.fa';

Readonly my $D_SUFFIX_OF_SEQ_FILES                  => 'seq.txt'; 
            # The suffix illumina uses to mark sequence files in Bustard
Readonly our $D_SUFFIX_OF_INT_FILES                  => 'int.txt';
            # The suffix illumina uses to mark intensity files in Firecrest
Readonly our $D_SYMBOL                               => '_' ; 
            #The symbol illumina uses to seperate field in files. 
            #For instance '_' in the case of: S_8_0001_seq.txt
Readonly my $D_GZ_SUFFIX                             => ".gz";
Readonly my $D_GUNZIP                                => 'gunzip';
Readonly my $D_GZIP                                  => 'gzip';

#^^^^^^^^^^^^^^^^^Headers^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


############################################ MAIN FUNCTION ######################################
sub quick {
    my ($arg_ref) = @_;

    my $verbose           = $arg_ref->{verbose};            #verbosity
    my $path_to_seq_files = $arg_ref->{path_to_seq_files};
    my $path_to_int_files = $arg_ref->{path_to_int_files};
    my $trim              = $arg_ref->{trim};
    my $output_dir        = $arg_ref->{output_dir};
    my $back              = $arg_ref->{maximum_lines_for_training};
    my $lane              = $arg_ref->{lane};
    my $r_tiles           = $arg_ref->{tiles};
    my $zipped            = $arg_ref->{zipped};



    #zipped
    my $suffix_int_files = $D_SUFFIX_OF_INT_FILES;
    if ($zipped) {
        $suffix_int_files .= $D_GZ_SUFFIX;
    }

    mkdir "$output_dir/$D_FOLDER" or confess "can't make folder: $D_FOLDER $!";
    $output_dir = "$output_dir/$D_FOLDER";

    my @seq_files_fullname = 
        get_filename_in_illumina_format({path   => $path_to_seq_files,
                                        lane    => $lane,
                                        tiles   => $r_tiles,
                                        suffix  => $D_SUFFIX_OF_SEQ_FILES,
                                        symbol  => $D_SYMBOL});
    my @int_files_fullname = 
        get_filename_in_illumina_format({path   => $path_to_int_files,
                                        lane    => $lane,
                                        tiles   =>$r_tiles,
                                        suffix  => $suffix_int_files,
                                        symbol  => $D_SYMBOL});


    my ($uni_seq_file,$uni_int_file) = 
        take_snapshot({
                       verbose=>$verbose,
                       back=> $back,
                       output_dir=>$output_dir,
                       seq_files=>\@seq_files_fullname,
                       int_files=>\@int_files_fullname,
                      });


    my ($uni_seq_to_blat_file, $unique_seq) = 
        prepare_to_blat({   
                         verbose=>$verbose,
                         unified_seq_sample=>$uni_seq_file,
                         output_dir=>$output_dir,
                         trim=>$trim,
                        });


    return ($uni_seq_to_blat_file,$uni_int_file);
}



############################# TAKE A SANPSHOT #######################################################3
sub take_snapshot {
    #this function samples random <back> lines from the training library - from the sequence files and the 
    # intensity files. The random lines are written in two files, one for seq and one for int. The two files
    # have the same content BUT the order is different. 
    # The function returns the fullnames of the two files as array. The first element is the seq file and the 
    # second is the int file.
    # NOTICE: Ns are not filetered by the function.

    my ($arg_ref) = shift;

    my $r_seq_files_to_read     =   $arg_ref->{'seq_files'};            #should be fullname files
    my $r_int_files_to_read     =   $arg_ref->{'int_files'};            #should be fullname files
    my $output_dir              =   $arg_ref->{'output_dir'} || '.';
    my $number_of_lines_back    =   $arg_ref->{'back'};                 #number of maximum lines to getback
    my $verbose                 =   $arg_ref->{'verbose'};


    # the output files are defined as constants:
    my @outputfiles = ("$output_dir/$D_NAME_OF_UNIFIED_SAMPLE_SEQ",
                       "$output_dir/$D_NAME_OF_UNIFIED_SAMPLE_INT");
                            #the order matters as we first process sequences files and than intensity files (see '**')
    my @copy_outputfiles = @outputfiles;
                            #copy_outputfiles is what returns from the function

    print "Taking sanpshot of the training library\n"
        if ($verbose);


    #how many lines do we have in total in the trainings library (on the tiles that were given)?
    my $total_lines = 
             total_number_of_lines_for_training({
                                                 files     => $r_seq_files_to_read,
                                                 verbose   => $verbose,
                                                 });

    #Scream if the requested lines back are more than the lines in the all library:
    confess "Total number of lines in the training library is less than",
          " requseted for sampling: $total_lines",
          " < $number_of_lines_back\n"
          if ($total_lines < $number_of_lines_back);


    # a reference to hash that the keys are line numbers (as read from the first file).
    # if the first file has 10000 lines and the second has 7000 - 11,000 referes to the 
    # 1000th line in the second file. The value is just 1 as it is used to check existsness.
    #
    # the function _genereate_random_sample_from return unique random <$number_of_lines_back> line numbers
    # from the interval 1-<$total_lines>.
    my $r_is_line_there     = 
            _genereate_random_sample_from($total_lines,
                                         $number_of_lines_back);

    #after we have the random lines numbers lets actual read them.
    # the SAMPLE_THEM loop is used to read and write first the sequence files and then the intensity files
    SAMPLE_THEM:
    foreach my $r_files ($r_seq_files_to_read,
                         $r_int_files_to_read) {

        # _sample reads the random lines from the files suffle and return
        # r_random_lines is array with all the lines
        my $r_random_lines = _sample($r_files,$r_is_line_there,$verbose);

        # (**) this is the reason the order matters:
        my $fileout = shift @outputfiles;

        #write the file:
        open my $OUT, '>', $fileout
            or confess "can't open $fileout: $!\n";

        print {$OUT} "$_" foreach (@$r_random_lines);

        close $OUT
            or confess "can't close $fileout: $!\n";
    } #end of SAMPLE_THEM

    print "Generated sample of sequences and intensities for training: @copy_outputfiles\n"
        if ($verbose);

    #return the fullnames of the written - (unified_seq,unifed_int).
    return @copy_outputfiles;
}




sub _sample {
    my ($r_files,$r_is_line_there,$verbose) = @_;

    #r_files = ref to array of files to read
    #r_is_line_there = hash function with line numbers
    #                  to sample as keys

    my $counter = 0;
    my @lines;
    my $gz_flag = 0;

    foreach my $file (@$r_files) {

        if ($file =~m/$D_GZ_SUFFIX/) {
            $file = _convert_gz_to_text_file($file);
            $gz_flag = 1;
        }

        open my $FH, '<', $file
            or confess "can't open $file: $!\n";

        print "Processing $file\n"
            if ($verbose);

        while (<$FH>) {
            $counter++;
            if (exists $r_is_line_there->{$counter}) {
                push @lines, $_;
            }
        }

        close $FH
            or confess "can't close $file: $!\n";

        if ($gz_flag == 1) {
            unlink $file || confess "can't clean $file: $!";
        }

    }

    @lines = shuffle @lines;
        # shuffle is for generating random order of training entries
        # the problem that solexa intensity files and seqeunce files
        #  are ordered by signal intensity.
    return \@lines;
}



sub _genereate_random_sample_from {
    my ($total_lines,$number_of_lines_back) = @_;

    #total_lines = sum of lines over all the files
    #number_of_lines_back = how many to sample out of them
    my @all_possible_lines = (1..$total_lines);
    @all_possible_lines = shuffle @all_possible_lines;
    my %is_line_there = map {$_ => '1'} @all_possible_lines[0..$number_of_lines_back-1];
    #return a hash with the line numbers as keys
    return \%is_line_there;
}


sub total_number_of_lines_for_training {
    #get list of files and return the total number of lines
    my ($arg_ref)           = shift;
    my $r_files_to_read     = $arg_ref->{'files'};
    my $verbose             = $arg_ref->{'verbose'}    || 0;
    my $total_lines         = 0;#counter

    confess "<files> should be array refernce\n"
        if (ref $r_files_to_read ne 'ARRAY');
    print "Reading training library\n"
        if ($verbose);

    READING_FILES:
    foreach my $file (@$r_files_to_read) {
        open my $FH, '<', $file
            or confess "can't open $file: $!\n";
        $total_lines+= tr/\n/\n/ while sysread ($FH, $_, 2**20);
        close $FH
            or confess "can't close $file: $!\n";
    } #end of READING_FILES

    print "Found: $total_lines lines in total\n"
        if ($verbose);
    return $total_lines;
}


############################# Prepare to blat #################################################################################


sub prepare_to_blat {
    #############################
    #this function create a dataset for training seqeunce. The sequence are trimmed 
    #and collapsed to boost blat search. The output file is in fasta format for blat.
    #  The header of each sequence is sarting with leading '>' and ';' seperates 
    # between the location of entries with the same sequence. 
    # It returns the path/filename of the output sequences and number of unique sequences. 

    # NOTICE: SEQUENCES WITH 'N' OR '.' ARE DISCARDED!

    #the inputs of the function are in a hash table with the following keys: unified_seq_sample,
    # trim, output_dir and verbose. Only unified_seq_sample is mandatory.
    # unified_seq_sample - a fullname of tab delimited file with cluster position as the first four colums
    # trim - number of bases to keep for training from the starting point of the sequence
    ##############################

    #get the function parameters
    my ($arg_ref) = @_;

    my $unified_seq_sample  = $arg_ref->{unified_seq_sample};
    my $trim                = $arg_ref->{trim};
    my $output_dir          = $arg_ref->{output_dir} || '.';
    my $verbose             = $arg_ref->{verbose} || 0;


    confess "take_snapshot error\n" 
        if any {not defined $_} $unified_seq_sample;

      #set variables:
    my %similar;
            #similar is used to collpase similar sequence entries
            #it holds a list of cluster positions for each sequence
    my $output_filename = "$output_dir/$D_NAME_OF_UNIFIED_SAMPLE_SEQ_FOR_BLAT";
    my $unique_seq = 0;
            #number of unique sequences for blat

    #prepare the outfile that contains the sequences for training

    READ_SEQUENCES: 
    {
        my $count = 0;
        open my $UNI_SEQ, '<', $unified_seq_sample 
                    or confess "can't open $unified_seq_sample: $!\n";
        while (<$UNI_SEQ>) {
            chomp;
            my ($lane,$tile,$x,$y,$seq) = split /\s+/ , $_;
            next if ($seq =~m/\./ or $seq =~m/N/);
                #very important: training sequences never have ambiguity in their called bases!
            $count++;
            my $header = join '_', ($lane,$tile,$x,$y);

            if (defined $trim) {
                $seq = substr $seq, 0, $trim;
            }
            push @{$similar{$seq}} , $header;
        } # end of READ_SEQUENCES
        close $UNI_SEQ;

        print "Found $count lines without ambiguous bases ('N's or '.')\n"
            if ($verbose);

    }

    WRITE_SEQUENCE: 
    {
        open my $FA,'>', $output_filename 
                or confess "can't open $output_filename for writing: $!\n";
        foreach my $seq (keys %similar) { 
            my $header = join (';',@{$similar{$seq}});
            $unique_seq++;
            print {$FA} ">$header\n","$seq\n" 
                    or confess "can print into $output_filename: $!\n";
                    #output file in fasta format
        } #end of WRITE_SEQUENCE

        print "After processing found ",$unique_seq," unique entries for blat\n"; 

        close $FA
            or confess "cant close $output_filename: $!\n";
    }

    return ($output_filename, $unique_seq);
}


############################# get_filename_in_illumina_format #################################################################################




sub get_filename_in_illumina_format {
    ###############
    #this function generates filenames in solexa format. For example s_4_0002_seq.txt
    #it get in hash table the following keys: path, lane, tile, suffix and symbol.
    #lane and tile are required, default path is './' , default symbol is '_'.
    #path does not need '/' at the end, and symbol is used to concatanate. 
    #for example: get_filename_in_illumina_format({lane => 3, tile => 100, suffix => seq.txt}) 
    #will give ./s_3_0100_seq.txt
    ###############

    my ($arg_ref) = @_;

    my $path      = $arg_ref->{path} || '.';
    my $lane      = $arg_ref->{lane};
    my $r_tiles   = $arg_ref->{tiles};
    my $suffix    = $arg_ref->{suffix};
    my $symbol    = $arg_ref->{symbol} || '_';

    my @results;

    confess 'get_filename_in_illumina_format error\n'
        if any {not defined $_} $lane, $r_tiles ;
        #die if user didn't pass lane or tile

    confess 'get_filename_in_illumina_format error - tiles should be array reference\n'
        if (ref $r_tiles ne 'ARRAY');

    foreach my $tile (@$r_tiles) {
        #generating /path/s_<lane>_000<tile>_<suffix>. like ./s_8_0001_seq.txt
        my $tile_with_zeros = sprintf "%04d" , $tile ;
        my $file = join $symbol, 's', $lane, $tile_with_zeros;
        if (defined $suffix) {
            $file = join $symbol, $file, $suffix;
        }

        my $file_full = join '/', $path, $file;
        push @results, $file_full;
    }
    return @results;
}

# 
# sub any (&@) {
#     my $f = shift;
#     return if ! @_;
#     for (@_) {
# 	return 1 if $f->();
#     }
#     return 0;
# }


################################## test and a template #####################################################

sub test {

    my @tiles = (1..4);
    my $verbose = 1;
    my $path_to_seq_files = '/mnt/mountPoint/usb1/071215_BILLIEHOLIDAY_0001_FC2020WAAXX/Data/C1-50_Firecrest1.8.28_31-12-2007_sbsuser/Bustard1.8.28_31-12-2007_sbsuser';
    my $path_to_int_files = '/mnt/mountPoint/usb1/071215_BILLIEHOLIDAY_0001_FC2020WAAXX/Data/C1-50_Firecrest1.8.28_31-12-2007_sbsuser';
    my $trim = 25;
    my $output_dir = '.';
    my $lines_back = 10000;
    my $lane = 8;

   my ($uni_seq_to_blat_file,$uni_int_file)=quick({verbose=>$verbose,
                                                   path_to_seq_files=>$path_to_seq_files,
                                                   path_to_int_files=>$path_to_int_files,
                                                   trim=>$trim,
                                                   output_dir=>$output_dir,
                                                   maximum_lines_for_training=>$lines_back,
                                                   lane=>$lane,
                                                   tiles=>\@tiles,
                                                   });

}


#####################################################################################################


sub _convert_gz_to_text_file {
    my ($gz_file) = @_;
    my $gz_basename = basename $gz_file;
    copy($gz_file,"/tmp/$gz_basename") or confess "can't copy $gz_file: $!\n";
    $gz_file = "/tmp/$gz_basename";
    system "$D_GUNZIP $gz_file";
    sleep (5);
    $gz_file =~ s/$D_GZ_SUFFIX//;
    return $gz_file;
}



1


