#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)


use strict;
use warnings;
use Readonly;
use Cwd qw (abs_path);
use Archive::Tar;
use Carp;
use Storable;
use Term::ANSIColor qw(:constants);
use Pod::Usage;
use Time::localtime;
use Getopt::Long;
use List::MoreUtils qw (any);

use FindBin qw($Bin);
use lib "$Bin/Cluster";
use lib "$Bin";
use Snapshot;
use Reporter qw (report);
use Cluster::Blat;
use Deconv;


#./brave_heart.pl --number_of_cycles 50 --path_to_intensity_files /home2/hannon/erlich/SolexaNoiseImprovment/perlProgram/AnalitycalModel/predict/files/ --path_to_training_sequences /home2/hannon/erlich/SGE-API/sequences_for_training_Classic/  --lane_for_training 8 --tiles_for_training 1-2 --reference_genome ./tmp/phiX.2bit --is_reference_genome_local Y --output_dir . --verbose 1 --trim 25 --threshold_for_analysis 20 --maximum_lines_for_training 50000


###########################################################################################
    $Term::ANSIColor::AUTORESET = 1;

    Readonly my $D_TRIM                                     =>          '25';
            #this number should be bigger than number of sequences
    Readonly my $D_LINES_BACK                               =>          '100000';
            #this number should be bigger than all $D_NUMBER_OF_SEQUENCES_FOR_ALL_CYCLES,
            #$D_INIT_C_FOR_P1_P2_OPT*$D_ROUNDS_OF_OPT_RANDOM_WALK
            #$D_NUMBER_OF_SEQUENCES_FOR_OPT_SVM

    Readonly my $D_THRESHOLD_FOR_ANALYSIS                   =>          '20';

    Readonly my $D_LIFETIME_FOR_OPT                         =>          '6h'; # NOTICE:10h former
    Readonly my $D_LIFETIME_FOR_ALL_CYCLE                   =>          '24h';
    Readonly my $D_REFRESH_TIME                             =>          '2m';
    Readonly my $D_ANALYSIS_THRESH                          =>          '25';

    Readonly my $D_ROUNDS_OF_OPT_RANDOM_WALK                =>          '3'; # NOTICE: 4 former
    Readonly my $D_MIDDLE_OF_P1_RANGE                       =>          '0.98';
    Readonly my $D_MIDDLE_OF_P2_RANGE                       =>          '0.98';
    Readonly my $D_HALF_RANGE_OF_P1_P2                      =>          '0.02';
    Readonly my $D_DIVIDER_OF_HALF_RANGE_OF_P1_P2           =>          '4'; # NOTICE 4 former
            #2*$D_HALF_RANGE_OF_P1_P2/($D_DIVIDER_OF_HALF_RANGE_OF_P1_P2 
            #** $D_ROUNDS_OF_OPT_RANDOM_WALK) is resolution!!!
    Readonly my $D_INIT_C_FOR_P1_P2_OPT                     =>          '1';
    Readonly my $D_INIT_NUMBER_OF_SEQUENCES_FOR_OPT_P1_P2   =>          '10000';
    Readonly my $D_INIT_GAMMA_FOR_P1_P2_OPT                 =>          1/$D_INIT_NUMBER_OF_SEQUENCES_FOR_OPT_P1_P2;

    Readonly my $D_ROUNDS_OF_OPT_SVM                        =>          '2';# NOTICE: 3 former
    Readonly my $D_MIDDLE_OF_C_RANGE                        =>          '5'; 
            #actually this is 2**$D_MIDDLE_OF_C_RANGE
    Readonly my $D_MIDDLE_OF_GAMMA_RANGE                    =>          '-6';
            #actually this is 2**$D_MIDDLE_OF_GAMMA_RANGE
    Readonly my $D_HALF_RANGE_OF_SVM                        =>          '8';
    Readonly my $D_DIVIDER_OF_HALF_RANGE_OF_SVM             =>          '4';
    Readonly my $D_NUMBER_OF_SEQUENCES_FOR_OPT_SVM          =>          '75000';

    Readonly my $D_NUMBER_OF_SEQUENCES_FOR_ALL_CYCLES       =>          '75000';

    Readonly my $D_SUFFIX_OF_HASH                           =>           "hash";


#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Headers ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
my $r_config = read_user_parameters();

load_hard_coded_parameters      ($r_config);
report_parameters               ($r_config);
blat_it                         ($r_config);
optimize_random_walk_parameters ($r_config);
optimize_SVM_parameters         ($r_config);
train_caller_for_all_cycles     ($r_config);
pack_files_for_karma            ($r_config);
report_optimization             ($r_config);




###################################### READ PARAMETERS ####################################################

sub read_user_parameters {


    my %config;


    my $number_of_cycles;
    my $path_to_intensity_files;
    my $path_to_training_sequences;
    my $lane_for_training;
    my $tiles_for_training;
    my $reference_genome;
    my $is_reference_genome_local;
    my $output_dir;
    my $trim;
    my $verbose;
    my $threshold_for_analysis;
    my $lines_back;
    my $debug;
    my $genome_locality;
    my $flowcellname;
    my $not_nice;
    my $zipped;


    my $tar_package_template;
    my $tar_package;
    my $param_file;


    my $HELP   = 0;
    my $MANUAL = 0;





    GetOptions( "number_of_cycles=i"                => \$number_of_cycles,
                "path_to_intensity_files=s"         => \$path_to_intensity_files,
                "path_to_training_sequences=s"      => \$path_to_training_sequences,
                "lane_for_training=i"               => \$lane_for_training,
                "tiles_for_training=s"              => \$tiles_for_training,
                "reference_genome=s"                => \$reference_genome,
                "is_reference_genome_local=s"       => \$is_reference_genome_local,
                "verbose"                           => \$verbose,
                "trim=i"                            =>\$trim,
                "threshold_for_analysis=i"          =>\$threshold_for_analysis,
                "maximum_lines_for_training=i"      =>\$lines_back,
                "debug"                             =>\$debug,
                "flowcell=s"                        =>\$flowcellname,
                "not_nice"                          =>\$not_nice,
                "zipped"                            => \$zipped,


                "help"                              =>\$HELP,
                "man"                               =>\$MANUAL,
                );

    pod2usage(1) if $HELP;
    pod2usage(-verbose => 2 ) if $MANUAL;


    if (
        (not defined $number_of_cycles)                 or
        (not defined $path_to_intensity_files)          or
        (not defined $lane_for_training)                or
        (not defined $tiles_for_training)               or
        (not defined $reference_genome)                 or
        (not defined $is_reference_genome_local)
       )                                                    {
        print "Not all mandatory parameters were loaded\n";
        pod2usage(1);
    }



    #genome locality
    if (($is_reference_genome_local eq 'Y') or ($is_reference_genome_local eq 'y')) {
        $genome_locality = 'local';
    }
    elsif (($is_reference_genome_local eq 'N') or ($is_reference_genome_local eq 'n')) {
        $genome_locality = 'distal';
    }

    #make path absolute
    if (not -d $path_to_intensity_files)    {confess "Can't find path_to_intensity_files: $path_to_intensity_files"}
    if (not -d $path_to_training_sequences) {confess "Can't find path_to_training_sequences: $path_to_training_sequences"}
    if (not -e $reference_genome and $genome_locality eq 'local') {confess "Can't find reference_genome: $reference_genome"}

    #
    $path_to_intensity_files    = abs_path $path_to_intensity_files;
    $path_to_training_sequences = abs_path $path_to_training_sequences;
    $reference_genome           = abs_path $reference_genome;




    if (not defined $trim) {
        $trim = $D_TRIM;
    }

    #tiles
    my ($start_tile,$stop_tile) = split /-/ , $tiles_for_training;
    if (not defined $stop_tile) {$stop_tile = $start_tile};
    my @tiles = ($start_tile..$stop_tile);






    #output_dir
    if (not defined $output_dir) {
        $output_dir = '.'; #this method is not working
    }
    ($output_dir,$tar_package_template) 
            = _auto_output_lib($output_dir,$flowcellname);

    chdir $output_dir or confess "can't change directory to $output_dir";

    #tar package
    $tar_package = $tar_package_template .'.Training_Files.tar.gz';
    $param_file  = $tar_package_template .'.Training_Param.param';

    #threshold_for_analysis
    if (not defined $threshold_for_analysis) {
        $threshold_for_analysis = $D_THRESHOLD_FOR_ANALYSIS;
    }

    #number of lines back
    if (not defined $lines_back) {
        $lines_back  =   $D_LINES_BACK;
    }



    $config{'number_of_cycles'}                 =       $number_of_cycles                       ;
    $config{'path_to_intensity_files'}          =       $path_to_intensity_files                ;
    $config{'path_to_training_sequences'}       =       $path_to_training_sequences             ;
    $config{'lane_for_training'}                =       $lane_for_training                      ;
    @{$config{'tiles_for_training_a'}}          =       @tiles                                  ;
    $config{'reference_genome'}                 =       $reference_genome                       ;
    $config{'is_reference_genome_local'}        =       $genome_locality                        ;
    $config{'output_dir'}                       =       $output_dir                             ;
    $config{'verbose'}                          =       $verbose                                ;
    $config{'trim'}                             =       $trim                                   ;
    $config{'threshold_for_analysis'}           =       $threshold_for_analysis                 ;
    $config{'lines_back'}                       =       $lines_back                             ;
    $config{'debug'}                            =       $debug                                  ;
    $config{'not_nice'}                         =       $not_nice                               ;
    $config{'zipped'}                           =       $zipped                                 ;
    $config{'flowcellname'}                     =       $flowcellname                           ;

    $config{'train_param_name'}                 =       $param_file                             ;
    $config{'train_tar_name'}                   =       $tar_package                            ; 

    return \%config;
}


################################# HARD CODED PARAMETERS ##########################################

sub load_hard_coded_parameters {

    my ($r_config) = @_;

    $r_config->{'threshold_for_analysis'}       =   $D_ANALYSIS_THRESH;



    $r_config->{'rounds_opt_random_walk'}       =   $D_ROUNDS_OF_OPT_RANDOM_WALK;
    $r_config->{'starting_best_p1'}             =   $D_MIDDLE_OF_P1_RANGE;
    $r_config->{'starting_best_p2'}             =   $D_MIDDLE_OF_P2_RANGE;
    $r_config->{'starting_gap_rand_opt'}        =   $D_HALF_RANGE_OF_P1_P2;
    $r_config->{'gap_div_rand_opt'}             =   $D_DIVIDER_OF_HALF_RANGE_OF_P1_P2; 
    $r_config->{'initial_c'}                    =   $D_INIT_C_FOR_P1_P2_OPT;
    $r_config->{'counter_thresh_for_rand_opt'}  =   $D_INIT_NUMBER_OF_SEQUENCES_FOR_OPT_P1_P2;
    $r_config->{'initial_gamma'}                =   $D_INIT_GAMMA_FOR_P1_P2_OPT;


    $r_config->{'rounds_opt_SVM'}               =   $D_ROUNDS_OF_OPT_SVM;
    $r_config->{'starting_best_c'}              =   $D_MIDDLE_OF_C_RANGE;
    $r_config->{'starting_best_gamma'}          =   $D_MIDDLE_OF_GAMMA_RANGE;
    $r_config->{'starting_gap_SVM_opt'}         =   $D_HALF_RANGE_OF_SVM;
    $r_config->{'gap_div_SVM_opt'}              =   $D_DIVIDER_OF_HALF_RANGE_OF_SVM;
    $r_config->{'counter_thresh_for_SVM_opt'}   =   $D_NUMBER_OF_SEQUENCES_FOR_OPT_SVM;
    $r_config->{'counter_thresh_for_all_c'}     =   $D_NUMBER_OF_SEQUENCES_FOR_ALL_CYCLES;


    # RANDOM
    $r_config->{'life_time_for_rand_opt'}       =   $D_LIFETIME_FOR_OPT;
    $r_config->{'refresh_time_for_rand_opt'}    =   $D_REFRESH_TIME;
    # SVM
    $r_config->{'life_time_for_SVM_opt'}        =   $D_LIFETIME_FOR_OPT;
    $r_config->{'refresh_time_for_SVM_opt'}     =   $D_REFRESH_TIME;
    # All Cycles
    $r_config->{'life_time_for_all_c'}          =   $D_LIFETIME_FOR_ALL_CYCLE;
    $r_config->{'refresh_time_for_all_c'}       =   $D_REFRESH_TIME;


    @{$r_config->{'training_cycles_for_rand_opt'}} = 
            (int ($r_config->{'number_of_cycles'}*0.9) ..$r_config->{'number_of_cycles'});


    for (my $i=1; $i<= $r_config->{'number_of_cycles'}; $i+=5) {
         push @{$r_config->{'training_cycles_for_SVM_opt'}}, $i;
    }

    return 1;
}




###################################### BLAT IT ####################################################
sub blat_it {

    my ($r_config) = @_;

    my $number_of_cycles            =   $r_config->{'number_of_cycles'}                     ;
    my $path_to_intensity_files     =   $r_config->{'path_to_intensity_files'}              ;
    my $path_to_training_sequences  =   $r_config->{'path_to_training_sequences'}           ;
    my $lane_for_training           =   $r_config->{'lane_for_training'}                    ;
    my @tiles_for_training_a        =   @{$r_config->{'tiles_for_training_a'}}              ;
    my $reference_genome            =   $r_config->{'reference_genome'}                     ;
    my $is_reference_genome_local   =   $r_config->{'is_reference_genome_local'}            ;
    my $output_dir                  =   $r_config->{'output_dir'}                           ;
    my $verbose                     =   $r_config->{'verbose'}                              ;
    my $trim                        =   $r_config->{'trim'}                                 ;
    my $threshold_for_analysis      =   $r_config->{'threshold_for_analysis'}               ;
    my $debug                       =   $r_config->{'debug'}                                ;
    my $lines_back                  =   $r_config->{'lines_back'}                           ;
    my $zipped                      =   $r_config->{'zipped'}                               ;

    #____________________________________________________________________________________
    confess "error in blat it"
    if any {not defined $_}

                            $number_of_cycles          ,$path_to_intensity_files,
                            $path_to_training_sequences,$lane_for_training,
                            $reference_genome          ,$is_reference_genome_local,
                            $output_dir                ,$trim,
                            $threshold_for_analysis    ,$lines_back;



     my ($uni_seq_to_blat_file,$uni_int_file) = Snapshot::quick 
                                                            ({verbose=>$verbose,
                                                             path_to_seq_files=>$path_to_training_sequences,
                                                             path_to_int_files=>$path_to_intensity_files,
                                                             trim=>$trim,
                                                             output_dir=>$output_dir,
                                                             maximum_lines_for_training=>$lines_back,
                                                             lane=>$lane_for_training,
                                                             tiles=>\@tiles_for_training_a,
                                                             zipped=>$zipped,
                                                            });

    my ($pos_to_seq_file) = Cluster::Blat::quick ({
                                                    genome_reference_2bit => $reference_genome,
                                                    genome_locality       => $is_reference_genome_local,
                                                    file_to_blat          => $uni_seq_to_blat_file,
                                                    dots                  => 100,
                                                    verbose               => $verbose,
                                                    threshold_for_analysis=> $threshold_for_analysis,
                                                    number_of_cycles      => $number_of_cycles,
                                                    minScore              => $threshold_for_analysis,
                                                    debug                 => $debug,
                                                   });


    $r_config->{'pos_to_seq_file'}      = $pos_to_seq_file;
    $r_config->{'uni_seq_to_blat_file'} = $uni_seq_to_blat_file;
    $r_config->{'uni_int_file'}         = $uni_int_file;
    return 1;
}






######################### Optimization of random walk parameters ##########################################
sub optimize_random_walk_parameters {
    my ($r_config) = @_;

    my $number_of_cycles            =   $r_config->{'number_of_cycles'}                     ;
    my $output_dir                  =   $r_config->{'output_dir'}                           ;
    my $verbose                     =   $r_config->{'verbose'}                              ;
    my $debug                       =   $r_config->{'debug'}                                ;

    my $pos_to_seq_file             =   $r_config->{'pos_to_seq_file'}                      ;
    my $uni_seq_to_blat_file        =   $r_config->{'uni_seq_to_blat_file'}                 ;
    my $uni_int_file                =   $r_config->{'uni_int_file'}                         ;


    my $rounds_opt_random_walk      =   $r_config->{'rounds_opt_random_walk'}               ;
    my $best_p1                     =   $r_config->{'starting_best_p1'}                     ;
    my $best_p2                     =   $r_config->{'starting_best_p2'}                     ;
    my $gap                         =   $r_config->{'starting_gap_rand_opt'}                ;
    my $gap_div                     =   $r_config->{'gap_div_rand_opt'}                     ;
    my $initial_c                   =   $r_config->{'initial_c'}                            ;
    my $initial_gamma               =   $r_config->{'initial_gamma'}                        ;
    my @training_cycles             =   @{$r_config->{'training_cycles_for_rand_opt'}}      ;

    my $counter_thresh_for_rand_opt =   $r_config->{'counter_thresh_for_rand_opt'}          ;
    my $life_time_for_rand_opt      =   $r_config->{'life_time_for_rand_opt'}               ;
    my $refresh_time_for_rand_opt   =   $r_config->{'refresh_time_for_rand_opt'}            ;



    #____________________________________________________________________________________
    confess "error in optimize_random_walk_parameters"
    if any {not defined $_}

                            $number_of_cycles          ,$output_dir,
                            $pos_to_seq_file           ,$uni_seq_to_blat_file,
                            $uni_int_file              ,$rounds_opt_random_walk,
                            $best_p1                   ,$best_p2,
                            $gap                       ,$gap_div,
                            $initial_c                 ,$initial_gamma,
                            $counter_thresh_for_rand_opt,$life_time_for_rand_opt,
                            $refresh_time_for_rand_opt
                        ;




    foreach my $round (1..$rounds_opt_random_walk) {

        my $deconv = Deconv->new();
        $deconv->set_c_g($initial_c,$initial_gamma);


        my $initial_point_p1 = $best_p1 -  $gap;
        my $initial_point_p2 = $best_p2 -  $gap;
        my $end_point_p1     = $best_p1 +  $gap;
        my $end_point_p2     = $best_p2 +  $gap;

        if ($end_point_p1 > 1) {$end_point_p1 = 1}
        if ($end_point_p2 > 1) {$end_point_p2 = 1}

        $gap = $gap/$gap_div;


        if ($verbose) {
            print "Optimization of random walk parameters:\n";
            print "Round $round\n";
            print "p1 = ($initial_point_p1:$gap:$end_point_p1)\n";
            print "p2 = ($initial_point_p2:$gap:$end_point_p2)\n";
        }

        my @p1 = ($initial_point_p1,$gap,$end_point_p1);
        my @p2 = ($initial_point_p2,$gap,$end_point_p2);

        if ($verbose) {
            $deconv->verbose;
        }
        if ($debug) {
            $deconv->flag_debug_mode_set;
        }


        $deconv->set_number_of_cycles($number_of_cycles);
        $deconv->set_counter_threshold($counter_thresh_for_rand_opt*$round);#!!!!!!!!!!
        $deconv->set_train_cycles(@training_cycles);
        $deconv->set_random_walk_range(\@p1,\@p2);
        $deconv->set_file_pos_to_seq($pos_to_seq_file);
        $deconv->set_unify_int_file($uni_int_file);
        $deconv->set_cross_val(2);
        $deconv->test_preparation_all;
        $deconv->generate_command_line;

        if ($debug) {
            $deconv->debug_print_parameters;
            $deconv->debug_print_command_line;
        }

        $deconv->qsub_all;
        $deconv->submit_life_time_set($life_time_for_rand_opt);
        $deconv->submit_quicky($refresh_time_for_rand_opt);
        $deconv->qsub_bye;

        $deconv->stdout_parse_matrix_quickly;
        if ($debug) {
            $deconv->debug_print_all_matrices;
            $deconv->debug_print_maximum_for_each_martix;
        }
        ($best_p1,$best_p2,undef) =  $deconv->matanl_maximum_upon_average ;
        $deconv->DESTROY;
        push @{$r_config->{optimization_step_p1p2}}, "[$best_p1,$best_p2]";
    }

    $r_config->{'best_p1'} = $best_p1;
    $r_config->{'best_p2'} = $best_p2;

    return 1;
}




######################### Optimization of C and gamma parameters ##########################################
sub optimize_SVM_parameters {

    my ($r_config) = @_;

    my $number_of_cycles            =   $r_config->{'number_of_cycles'}                     ;
    my $output_dir                  =   $r_config->{'output_dir'}                           ;
    my $verbose                     =   $r_config->{'verbose'}                              ;
    my $trim                        =   $r_config->{'trim'}                                 ;
    my $debug                       =   $r_config->{'debug'}                                ;

    my $pos_to_seq_file             =   $r_config->{'pos_to_seq_file'}                      ;
    my $uni_seq_to_blat_file        =   $r_config->{'uni_seq_to_blat_file'}                 ;
    my $uni_int_file                =   $r_config->{'uni_int_file'}                         ;


    my $rounds_opt_SVM              =   $r_config->{'rounds_opt_SVM'}                       ;
    my $best_c                      =   $r_config->{'starting_best_c'}                      ;
    my $best_gamma                  =   $r_config->{'starting_best_gamma'}                  ;
    my $gap                         =   $r_config->{'starting_gap_SVM_opt'}                 ;
    my $gap_div                     =   $r_config->{'gap_div_SVM_opt'}                      ;
    my $best_p1                     =   $r_config->{'best_p1'}                              ;
    my $best_p2                     =   $r_config->{'best_p2'}                              ;
    my @training_cycles             =   @{$r_config->{'training_cycles_for_SVM_opt'}}       ;

    my $counter_thresh_for_SVM_opt  =   $r_config->{'counter_thresh_for_SVM_opt'}           ;
    my $life_time_for_SVM_opt       =   $r_config->{'life_time_for_SVM_opt'}                ;
    my $refresh_time_for_SVM_opt    =   $r_config->{'refresh_time_for_SVM_opt'}             ;

    #____________________________________________________________________________________

    confess "error in optimize_SVM_parameters"
    if any {not defined $_}

                            $number_of_cycles          ,$output_dir,
                            $pos_to_seq_file           ,$uni_seq_to_blat_file,
                            $uni_int_file              ,$rounds_opt_SVM,
                            $best_c                    ,$best_gamma,
                            $gap                       ,$gap_div,
                            $best_p1                   ,$best_p2,
                            $counter_thresh_for_SVM_opt,$life_time_for_SVM_opt,
                            $refresh_time_for_SVM_opt
                        ;


 
    foreach my $round (1..$rounds_opt_SVM) {

        my $deconv = Deconv->new();

        my $initial_point_c     = $best_c     -  $gap;
        my $end_point_c         = $best_c     +  $gap;
        my $initial_point_gamma = $best_gamma -  $gap;
        my $end_point_gamma     = $best_gamma +  $gap;

        $gap = $gap/$gap_div;

        if ($verbose) {
            print "Optimization of SVM parameters:\n";
            print "Round $round\n";
            print "c = ($initial_point_c:$gap:$end_point_c)\n";
            print "gamma = ($initial_point_gamma:$gap:$end_point_gamma)\n";
        }

        my @c = ($initial_point_c,$gap,$end_point_c);
        my @gamma = ($initial_point_gamma,$gap,$end_point_gamma);

        if ($verbose) {
            $deconv->verbose;
        }

        if ($debug) {
            $deconv->flag_debug_mode_set;
        }

        $deconv->set_svm_range(\@c,\@gamma);
        $deconv->set_p1_p2($best_p1,$best_p2);
        $deconv->set_train_cycles(@training_cycles);

        $deconv->set_number_of_cycles($number_of_cycles);
        $deconv->set_counter_threshold($counter_thresh_for_SVM_opt);
        $deconv->set_file_pos_to_seq($pos_to_seq_file);
        $deconv->set_unify_int_file($uni_int_file);
        $deconv->set_cross_val(2);

        $deconv->test_preparation_all;
        $deconv->generate_command_line;

        if ($debug) {
            $deconv->debug_print_parameters;
            $deconv->debug_print_command_line;
        }

        $deconv->qsub_all;
        $deconv->submit_life_time_set($life_time_for_SVM_opt);
        $deconv->submit_quicky($refresh_time_for_SVM_opt);
        $deconv->qsub_bye;

        $deconv->stdout_parse_matrix_quickly;
        if ($debug) {
            $deconv->debug_print_all_matrices;
            $deconv->debug_print_maximum_for_each_martix;
        }
        ($best_c,$best_gamma,undef) =  $deconv->matanl_maximum_upon_average ;

        push @{$r_config->{optimization_step_c_gamma}}, "[$best_c,$best_gamma]";
# 
        $deconv->DESTROY;

    }

   $r_config->{'best_c'}     = $best_c;
   $r_config->{'best_gamma'} = $best_gamma;

    return 1;
}




######################### Train Caller for all cycles ##########################################
sub train_caller_for_all_cycles {

    my ($r_config) = @_;
    my $number_of_cycles            =   $r_config->{'number_of_cycles'}                     ;
    my $output_dir                  =   $r_config->{'output_dir'}                           ;
    my $verbose                     =   $r_config->{'verbose'}                              ;
    my $trim                        =   $r_config->{'trim'}                                 ;
    my $debug                       =   $r_config->{'debug'}                                ;

    my $pos_to_seq_file             =   $r_config->{'pos_to_seq_file'}                      ;
    my $uni_seq_to_blat_file        =   $r_config->{'uni_seq_to_blat_file'}                 ;
    my $uni_int_file                =   $r_config->{'uni_int_file'}                         ;


    my $best_c                      =   $r_config->{'best_c'}                               ;
    my $best_gamma                  =   $r_config->{'best_gamma'}                           ;
    my $best_p1                     =   $r_config->{'best_p1'}                              ;
    my $best_p2                     =   $r_config->{'best_p2'}                              ;

    my $counter_thresh_for_all_c    =   $r_config->{'counter_thresh_for_all_c'}             ;
    my $life_time_for_all_c         =   $r_config->{'life_time_for_all_c'}                  ;
    my $refresh_time_for_all_c      =   $r_config->{'refresh_time_for_all_c'}               ;

    #____________________________________________________________________________________
    confess "error in train_caller_for_all_cycles"
    if any {not defined $_}

                            $number_of_cycles          ,$output_dir,
                            $pos_to_seq_file           ,$uni_seq_to_blat_file,
                            $uni_int_file,
                            $best_c                    ,$best_gamma,
                            $best_p1                   ,$best_p2,
                            $counter_thresh_for_all_c  ,$life_time_for_all_c,
                            $refresh_time_for_all_c
                        ;

    my %cycle_to_file;
    my $deconv = Deconv->new();

    $deconv->set_cycle_train_range(1,1,$number_of_cycles);
    $deconv->set_estimate_prop;
    $deconv->set_counter_threshold($counter_thresh_for_all_c);
    $deconv->set_c_g($best_c,$best_gamma);
    $deconv->set_p1_p2($best_p1,$best_p2);
    $deconv->set_file_pos_to_seq($pos_to_seq_file);
    $deconv->set_unify_int_file($uni_int_file);
    $deconv->set_bring_back_files_to($output_dir);
    $deconv->set_number_of_cycles($number_of_cycles);

    $deconv->test_preparation_all;
    $deconv->generate_command_line;

    if ($debug) {
        $deconv->flag_debug_mode_set;
        $deconv->debug_print_parameters;
        $deconv->debug_print_command_line;
    }

    $deconv->qsub_all;
    $deconv->submit_life_time_set($life_time_for_all_c);
    $deconv->submit_quicky($refresh_time_for_all_c);
    $deconv->qsub_bye;
    $deconv->stdout_parse_returned_files_quickly;

    ARRANGE_RETURNERED_FILES:
    foreach my $cycle (1..$number_of_cycles) {
        my %findMe = (train_cycles=>$cycle);
        my $file = $deconv->stdout_looktable__search(\%findMe);
        $cycle_to_file{$cycle} = $file;
        if ($debug) {
            print "$cycle\t$file\n";
        }
    }
    $deconv->DESTROY;


    $r_config->{'r_hash_cycle_to_file'} = \%cycle_to_file;
    return 1;
}



######################### Pack files ##########################################

sub pack_files_for_karma {

    my ($r_config) = @_;

    my $verbose                     =   $r_config->{'verbose'}                              ;
    my $train_tar_name              =   $r_config->{'train_tar_name'}                       ;
    my $r_hash_cycle_to_file        =   $r_config->{'r_hash_cycle_to_file'}                 ;
    my $param_file                  =   $r_config->{'train_param_name'}                     ;
    my $not_nice                    =   $r_config->{'not_nice'}                             ;

    my @files_to_pack;

    #____________________________________________________________________________________



    foreach my $cycle (sort {$a <=> $b} keys %$r_hash_cycle_to_file) {
        my $file = $r_hash_cycle_to_file->{$cycle};
        my $tar = Archive::Tar->new();
        my $read = $tar->read($file,1);
        my @files_to_take = grep {$_ =~ m/cycle_$cycle\_/} $tar->list_files;
        confess "Problem taking files. Found: @files_to_take"
            if (
                (scalar @files_to_take != 2)                             or
                ((scalar (grep {$_ =~ m/\.model/} @files_to_take)) == 0) or
                ((scalar (grep {$_ =~ m/\.range/} @files_to_take)) == 0)
                );

        $tar->extract(@files_to_take);
        push @files_to_pack,@files_to_take;
        unlink $file or confess "can't clean $file: $!";
    }

    if ($verbose) {
        print "$_\n" foreach (@files_to_pack);
    }




    my $tar = Archive::Tar->new();
    $tar->add_files(@files_to_pack);
    my $filename = $train_tar_name;


    # Write outputs
    if (not -d "Karma") {
        mkdir "Karma" || confess "can't mkdir: $!";
    }
    chdir "Karma";
    #write tar
    $tar->write($filename,1);
    $r_config->{'SVM_tar_location'} = abs_path $filename;

    #write hash_for_human
    open my $FH, '>', $param_file 
        || confess "can't open $param_file: $!";
    _print_hash_to_file($r_config,$FH);
    close $FH;

    #write hash
    my $file_hash = $param_file.".$D_SUFFIX_OF_HASH";
    store($r_config,$file_hash);
    chdir "..";


    if (not $not_nice) {
        Reporter::report($r_config);
    }

    foreach my $file (@files_to_pack) {
        unlink $file or confess "can't clean $file: $!\n";
    }

    system ("touch Training.finished");
    return 1;
}










######################### Aux ##########################################
sub _auto_output_lib {
    my $user_dir  = shift;
    my $flowcellname = shift || '';

    $user_dir =~ s/\/$//; # get rid of the last / if present in the dir

    if (not -d $user_dir) {
        mkdir $user_dir ||
            confess "can't mkdir $user_dir: $!\n";
    }
    my $tm = localtime;
    my $folderName = sprintf("%04d.%02d.%02d_%02d.%02d.%02d__Training.Folder_%s",
                             ($tm->year)+1900      ,($tm->mon)+1,
                              $tm->mday             ,$tm->hour,
                              $tm->min              ,$tm->sec,
                              $flowcellname);

    my $folderName_full = "$user_dir/$folderName";

    my $copy_folderName_full = $folderName_full;
    my $count = 1;
    while (-d $copy_folderName_full) {
        $count++;
        $copy_folderName_full = "$folderName_full.$count";
    }

    mkdir $copy_folderName_full ||
        confess "can't mkdir $copy_folderName_full: $!\n";

    $copy_folderName_full = abs_path $copy_folderName_full;

        #folderName is used later for Tar name;
    return ($copy_folderName_full,$folderName);
}

sub report_parameters {
    my ($r_config) = @_;
    if ($r_config->{'verbose'}) {
    print "____________________________________________________________________________________\n";
    print "Training paramters:\n";
    _print_hash ($r_config);
    print "____________________________________________________________________________________\n";
    }
    return 1;
}

sub report_optimization {
    my ($r_config) = @_;
    my @key_s = qw(best_p1 best_p2 best_c best_gamma);

    print "____________________________________________________________________________________\n";
    print "Optimization results:\n";
    foreach my $key (@key_s) {
        print BOLD, WHITE, "$key", RESET;
        my $l = 50 - length ($key);
        print " " foreach (1..$l);
        print $r_config->{$key},"\n";
    }
    print "____________________________________________________________________________________\n";
    return 1;
}


# ugly code
sub _print_hash {
    my ($r_hash) = @_;
    foreach my $k (sort keys %$r_hash) {
        my $v = $r_hash->{$k};
        if (not defined $v) {$v = 'undef'}
        print BOLD, WHITE, "$k", RESET;
        my $l = 50 - length ($k);
        print " " foreach (1..$l);
        if (ref $v eq 'ARRAY') {
            print "@$v\n";
        }
        else {
            print "$v\n";
        }
    }
}

sub _print_hash_to_file {
    my ($r_hash,$FH) = @_;
    foreach my $k (sort keys %$r_hash) {
        my $v = $r_hash->{$k};
        if (not defined $v) {$v = 'undef'}
        print {$FH} "$k";
        my $l = 50 - length ($k);
        print {$FH} " " foreach (1..$l);
        if (ref $v eq 'ARRAY') {
            print {$FH} "@$v\n";
        }
        else {
            print {$FH} "$v\n";
        }
    }
}
# end of ugly code

__END__

=head1 NAME


brave_heart - A trainer program for Alta-Cyclic base caller.

=head1 SYNOPSIS

brave_heart [MANDATORY PARAMETERS]   [OPTIONAL PARAMETERS] 

=cut


=head1 OPTIONS


  Mandatory Parameters:
  --number_of_cycles                Number of cycles in the run
  --path_to_intensity_files         Path to the folder that contains the intensity files for training
  --path_to_training_sequences      Path to the folder that contains sequences that correspond to training intensity files
  --lane_for_training               Lane number that is used for the training 
  --tiles_for_training              The range of tiles that will be used for training
  --reference_genome                Full path to the reference genome. The reference should be in 2bit format.
  --is_reference_genome_local       Y - indicates that the refernce genome should be copied to the execution nodes
                                    N - assumes that the genome is installed on the cluster nodes


  Optional Parameters:
  --trim                            Trim the trainings sequences to a certain size. Makes the training faster
  --threshold_for_analysis          Set the threshold for allowing mismatches in the training seqeunces
  --maximum_lines_for_training      Set the intial amount of eneries for sampling.
  --flowcell                        The flowcell name. 
  --verbose                         Set verbosity.
  --zipped                          If the intensity files are zipped
  --debug                           Give extra messages and indicators. Stores .o files that return from the cluster
  --not_nice                        Don't report your optimization results to Alta-Cyclic developers. This may prevent us of collecting statistics and improve later versions.

  Help Options:
   --help                           Show a list of parameters.
   --man                            Print a long manual.


=cut


=head1 EXAMPLES

brave_heart.pl --number_of_cycles 50 --path_to_intensity_files /home2/hannon/erlich/SGE-API/int_from_the_classic_run --path_to_training_sequences /home2/hannon/erlich/SGE-API/sequences_for_training_Classic/  --lane_for_training 8 --tiles_for_training 1-5 --reference_genome ../step2/tmp/phiX.2bit --is_reference_genome_local Y --trim 25 --threshold_for_analysis 20 --maximum_lines_for_training 100000 --verbose --debug --flowcell FLOWCELL_1234 --trim 25 --threshold_for_analysis 20 


=cut


=head1 DESCRIPTION

brave_heart.pl is part of Alta-Cyclic base caller. The program extracts a sample of intensity reads and their corresponding sequences that were called with standard illumina base caller.
The sequences are trimmed and matched back to the genome of the organism to extract a full length sequences. The genomic sequences and the intensity values form the training set. The training
set is first used for optimization of the random walk parameters (phasing) using iterative grid search. After optimization of the the random walk parameters, the program optimizes the SVM parameters (cross-talk).
Finally, an SVM machines are prodouced for every cycle. This can be later used for calling using karma.pl


=cut

=head1 BUGS AND LIMITATIONS

If you feel you have found a bug in brave_heart.pl, please submit a bug report at erlich@cshl.edu

=cut

=head1 AUTHOR

Written By Yaniv Erlich <erlich@cshl.edu>

=cut

=head1 LISCENSE AND COPYRIGHT

Copyright © 2008 Cold Spring Harbor Laboratory.
This is free software.  You may redistribute copies of it under the terms in LICENSE file.  There is NO WARRANTY, to the extent permitted by law.

=cut
